@@ -6,6 +6,7 @@ import (
 	"encoding/base64"
 	"encoding/json"
 	"fmt"
+	"io/ioutil"
 	"net/url"
 	"os"
 	"strings"
@@ -14,12 +15,17 @@ import (
 	apiextensionsv1beta1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
 	apiextensionsclient "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
 
+	addonv1alpha1 "github.com/open-cluster-management/api/addon/v1alpha1"
+	addonclient "github.com/open-cluster-management/api/client/addon/clientset/versioned"
 	workclient "github.com/open-cluster-management/api/client/work/clientset/versioned"
 	clusterv1 "github.com/open-cluster-management/api/cluster/v1"
 	workv1 "github.com/open-cluster-management/api/work/v1"
 	configv1alpha1 "github.com/open-cluster-management/submariner-addon/pkg/apis/submarinerconfig/v1alpha1"
 	configclient "github.com/open-cluster-management/submariner-addon/pkg/client/submarinerconfig/clientset/versioned"
 
+	submarinerv1alpha1 "github.com/submariner-io/submariner-operator/apis/submariner/v1alpha1"
+	submarinermv1 "github.com/submariner-io/submariner/pkg/apis/submariner.io/v1"
+
 	"github.com/openshift/api"
 	apicofigv1 "github.com/openshift/api/config/v1"
 	"github.com/openshift/library-go/pkg/operator/events"
@@ -44,7 +50,15 @@ import (
 )
 
 const (
-	ClusterTypeOCP       = "OCP"
+	SubmarinerAddOnName  = "submariner"
+	SubmarinerConfigName = "submariner"
+)
+
+const (
+	ProductOCP = "OpenShift"
+)
+
+const (
 	IPSecPSKSecretLength = 48
 	IPSecPSKSecretName   = "submariner-ipsec-psk"
 	BrokerAPIServer      = "BROKER_API_SERVER"
@@ -63,6 +77,16 @@ const (
 	ocpConfigNamespace    = "openshift-config"
 )
 
+const (
+	submarinerConnectionDegraded = "SubmarinerConnectionDegraded"
+)
+
+const (
+	brokerSuffix            = "broker"
+	namespaceMaxLength      = 63
+	clusterSetNameMaxLength = namespaceMaxLength - len(brokerSuffix) - 1
+)
+
 var (
 	infrastructureGVR = schema.GroupVersionResource{
 		Group:    "config.openshift.io",
@@ -144,6 +168,55 @@ func UpdateSubmarinerConfigStatusFn(cond metav1.Condition, managedClusterInfo co
 	}
 }
 
+type UpdateManagedClusterAddOnStatusFunc func(status *addonv1alpha1.ManagedClusterAddOnStatus) error
+
+func UpdateManagedClusterAddOnStatus(
+	ctx context.Context,
+	client addonclient.Interface,
+	addOnNamespace, addOnName string,
+	updateFuncs ...UpdateManagedClusterAddOnStatusFunc) (*addonv1alpha1.ManagedClusterAddOnStatus, bool, error) {
+	updated := false
+	var updatedAddOnStatus *addonv1alpha1.ManagedClusterAddOnStatus
+
+	err := retry.RetryOnConflict(retry.DefaultBackoff, func() error {
+		addOn, err := client.AddonV1alpha1().ManagedClusterAddOns(addOnNamespace).Get(ctx, addOnName, metav1.GetOptions{})
+		if err != nil {
+			return err
+		}
+		oldStatus := &addOn.Status
+
+		newStatus := oldStatus.DeepCopy()
+		for _, update := range updateFuncs {
+			if err := update(newStatus); err != nil {
+				return err
+			}
+		}
+		if equality.Semantic.DeepEqual(oldStatus, newStatus) {
+			// We return the newStatus which is a deep copy of oldStatus but with all update funcs applied.
+			updatedAddOnStatus = newStatus
+			return nil
+		}
+
+		addOn.Status = *newStatus
+		updatedAddOn, err := client.AddonV1alpha1().ManagedClusterAddOns(addOnNamespace).UpdateStatus(ctx, addOn, metav1.UpdateOptions{})
+		if err != nil {
+			return err
+		}
+		updatedAddOnStatus = &updatedAddOn.Status
+		updated = err == nil
+		return err
+	})
+
+	return updatedAddOnStatus, updated, err
+}
+
+func UpdateManagedClusterAddOnStatusFn(cond metav1.Condition) UpdateManagedClusterAddOnStatusFunc {
+	return func(oldStatus *addonv1alpha1.ManagedClusterAddOnStatus) error {
+		meta.SetStatusCondition(&oldStatus.Conditions, cond)
+		return nil
+	}
+}
+
 type CRDClientHolder struct {
 	apiExtensionsClient apiextensionsclient.Interface
 }
@@ -241,6 +314,7 @@ func CleanUpSubmarinerManifests(
 			errs = append(errs, err)
 			continue
 		}
+
 		gvk := resourcehelper.GuessObjectGroupVersionKind(object)
 		recorder.Eventf(fmt.Sprintf("Submariner%sDeleted", gvk.Kind), "Deleted %s", resourcehelper.FormatResourceForCLIWithNamespace(object))
 	}
@@ -338,13 +412,10 @@ func GetBrokerTokenAndCA(kubeClient kubernetes.Interface, dynamicClient dynamic.
 
 }
 
-func GetClusterType(managedCluster *clusterv1.ManagedCluster) string {
-	if clusterType, found := managedCluster.GetLabels()["vendor"]; found {
-		switch clusterType {
-		case "OCP", "OpenShift":
-			return ClusterTypeOCP
-		default:
-			return clusterType
+func GetClusterProduct(managedCluster *clusterv1.ManagedCluster) string {
+	for _, claim := range managedCluster.Status.ClusterClaims {
+		if claim.Name == "product.open-cluster-management.io" {
+			return claim.Value
 		}
 	}
 
@@ -359,34 +430,46 @@ func GetEnv(key, defaultValue string) string {
 	return value
 }
 
-func ApplyManifestWork(ctx context.Context, client workclient.Interface, required *workv1.ManifestWork) error {
-	err := retry.RetryOnConflict(retry.DefaultBackoff, func() error {
+func ApplyManifestWork(ctx context.Context, client workclient.Interface, required *workv1.ManifestWork, recorder events.Recorder) error {
+	return retry.RetryOnConflict(retry.DefaultBackoff, func() error {
 		existing, err := client.WorkV1().ManifestWorks(required.Namespace).Get(ctx, required.Name, metav1.GetOptions{})
-		if err != nil {
-			if errors.IsNotFound(err) {
-				_, err = client.WorkV1().ManifestWorks(required.Namespace).Create(ctx, required, metav1.CreateOptions{})
+		switch {
+		case errors.IsNotFound(err):
+			_, err := client.WorkV1().ManifestWorks(required.Namespace).Create(ctx, required, metav1.CreateOptions{})
+			if err != nil {
+				return err
 			}
+
+			recorder.Event("ManifestWorkApplied", fmt.Sprintf("manifestwork %s/%s was created", required.Namespace, required.Name))
+			return nil
+		case err != nil:
 			return err
 		}
 
-		if !manifestsEqual(existing.Spec.Workload.Manifests, required.Spec.Workload.Manifests) {
-			existingCopy := existing.DeepCopy()
-			existingCopy.Spec = required.Spec
-			_, err = client.WorkV1().ManifestWorks(required.Namespace).Update(ctx, existingCopy, metav1.UpdateOptions{})
+		if manifestsEqual(existing.Spec.Workload.Manifests, required.Spec.Workload.Manifests) {
+			return nil
 		}
 
-		return err
-	})
+		existingCopy := existing.DeepCopy()
+		existingCopy.Spec = required.Spec
+		_, err = client.WorkV1().ManifestWorks(required.Namespace).Update(ctx, existingCopy, metav1.UpdateOptions{})
+		if err != nil {
+			return err
+		}
 
-	return err
+		recorder.Event("ManifestWorkApplied", fmt.Sprintf("manifestwork %s/%s was updated", required.Namespace, required.Name))
+		return nil
+	})
 }
 
 func GetManagedClusterInfo(managedCluster *clusterv1.ManagedCluster) configv1alpha1.ManagedClusterInfo {
 	clusterInfo := configv1alpha1.ManagedClusterInfo{
 		ClusterName: managedCluster.Name,
-		Vendor:      GetClusterType(managedCluster),
 	}
 	for _, claim := range managedCluster.Status.ClusterClaims {
+		if claim.Name == "product.open-cluster-management.io" {
+			clusterInfo.Vendor = claim.Value
+		}
 		if claim.Name == "platform.open-cluster-management.io" {
 			clusterInfo.Platform = claim.Value
 		}
@@ -464,3 +547,108 @@ func getKubeAPIServerCA(kubeClient kubernetes.Interface, dynamicClient dynamic.I
 
 	return nil, nil
 }
+
+func CheckSubmarinerConnections(clusterName string, submariner *submarinerv1alpha1.Submariner) metav1.Condition {
+	condition := metav1.Condition{
+		Type: submarinerConnectionDegraded,
+	}
+	if submariner.Status.Gateways == nil || len(*submariner.Status.Gateways) == 0 {
+		condition.Status = metav1.ConditionTrue
+		condition.Reason = "ConnectionsNotEstablished"
+		condition.Message = "There are no connections on gateways"
+		return condition
+	}
+
+	connectedMessages := []string{}
+	unconnectedMessages := []string{}
+	for _, gateway := range *submariner.Status.Gateways {
+		for _, connection := range gateway.Connections {
+			if connection.Status != submarinermv1.Connected {
+				unconnectedMessages = append(unconnectedMessages, fmt.Sprintf("The connection between clusters %q and %q is not established (status=%s)",
+					clusterName, connection.Endpoint.ClusterID, connection.Status))
+				continue
+			}
+
+			connectedMessages = append(connectedMessages, fmt.Sprintf("The connection between clusters %q and %q is established",
+				clusterName, connection.Endpoint.ClusterID))
+		}
+	}
+
+	if len(connectedMessages) == 0 && len(unconnectedMessages) == 0 {
+		condition.Status = metav1.ConditionTrue
+		condition.Reason = "ConnectionsNotEstablished"
+		condition.Message = "There are no connections on gateways"
+		return condition
+	}
+
+	if len(unconnectedMessages) != 0 {
+		condition.Status = metav1.ConditionTrue
+		condition.Reason = "ConnectionsDegraded"
+		connectedMessages = append(connectedMessages, unconnectedMessages...)
+		condition.Message = strings.Join(connectedMessages, "\n")
+		return condition
+	}
+
+	condition.Status = metav1.ConditionFalse
+	condition.Reason = "ConnectionsEstablished"
+	condition.Message = strings.Join(connectedMessages, "\n")
+	return condition
+}
+
+// GetCurrentNamespace returns the current namesapce from file system,
+// if the namespace is not found, it returns the defaultNamespace
+func GetCurrentNamespace(defaultNamespace string) string {
+	nsBytes, err := ioutil.ReadFile("/var/run/secrets/kubernetes.io/serviceaccount/namespace")
+	if err != nil {
+		return defaultNamespace
+	}
+	return string(nsBytes)
+}
+
+func GernerateBrokerName(clusterSetName string) string {
+	name := fmt.Sprintf("%s-%s", clusterSetName, brokerSuffix)
+	if len(name) > namespaceMaxLength {
+		truncatedClusterSetName := clusterSetName[(len(brokerSuffix) - 1):]
+		return fmt.Sprintf("%s-%s", truncatedClusterSetName, brokerSuffix)
+	}
+	return name
+}
+
+func RemoveConfigFinalizer(ctx context.Context, configClient configclient.Interface, config *configv1alpha1.SubmarinerConfig, finalizer string) error {
+	copiedFinalizers := []string{}
+	for i := range config.Finalizers {
+		if config.Finalizers[i] == finalizer {
+			continue
+		}
+		copiedFinalizers = append(copiedFinalizers, config.Finalizers[i])
+	}
+
+	if len(config.Finalizers) != len(copiedFinalizers) {
+		copied := config.DeepCopy()
+		copied.Finalizers = copiedFinalizers
+		_, err := configClient.SubmarineraddonV1alpha1().SubmarinerConfigs(copied.Namespace).Update(ctx, copied, metav1.UpdateOptions{})
+		return err
+	}
+
+	return nil
+}
+
+// RemoveAddOnFinalizer removes the addon finalizer from a submariner-addon
+func RemoveAddOnFinalizer(ctx context.Context, addOnClient addonclient.Interface, addOn *addonv1alpha1.ManagedClusterAddOn, finalizer string) error {
+	copiedFinalizers := []string{}
+	for i := range addOn.Finalizers {
+		if addOn.Finalizers[i] == finalizer {
+			continue
+		}
+		copiedFinalizers = append(copiedFinalizers, addOn.Finalizers[i])
+	}
+
+	if len(addOn.Finalizers) != len(copiedFinalizers) {
+		copied := addOn.DeepCopy()
+		copied.Finalizers = copiedFinalizers
+		_, err := addOnClient.AddonV1alpha1().ManagedClusterAddOns(copied.Namespace).Update(ctx, copied, metav1.UpdateOptions{})
+		return err
+	}
+
+	return nil
+}
