@@ -7,6 +7,10 @@ import (
 
 	"github.com/ghodss/yaml"
 
+	addonv1alpha1 "github.com/open-cluster-management/api/addon/v1alpha1"
+	addonclient "github.com/open-cluster-management/api/client/addon/clientset/versioned"
+	addoninformerv1alpha1 "github.com/open-cluster-management/api/client/addon/informers/externalversions/addon/v1alpha1"
+	addonlisterv1alpha1 "github.com/open-cluster-management/api/client/addon/listers/addon/v1alpha1"
 	clusterclient "github.com/open-cluster-management/api/client/cluster/clientset/versioned"
 	clusterinformerv1 "github.com/open-cluster-management/api/client/cluster/informers/externalversions/cluster/v1"
 	clusterinformerv1alpha1 "github.com/open-cluster-management/api/client/cluster/informers/externalversions/cluster/v1alpha1"
@@ -20,6 +24,7 @@ import (
 	configv1alpha1 "github.com/open-cluster-management/submariner-addon/pkg/apis/submarinerconfig/v1alpha1"
 	configclient "github.com/open-cluster-management/submariner-addon/pkg/client/submarinerconfig/clientset/versioned"
 	configinformer "github.com/open-cluster-management/submariner-addon/pkg/client/submarinerconfig/informers/externalversions/submarinerconfig/v1alpha1"
+	configlister "github.com/open-cluster-management/submariner-addon/pkg/client/submarinerconfig/listers/submarinerconfig/v1alpha1"
 	"github.com/open-cluster-management/submariner-addon/pkg/cloud"
 	"github.com/open-cluster-management/submariner-addon/pkg/helpers"
 	"github.com/open-cluster-management/submariner-addon/pkg/hub/submarineragent/bindata"
@@ -42,13 +47,17 @@ import (
 	"k8s.io/klog/v2"
 )
 
+const manifestWorkName = "submariner-operator"
+
+const (
+	clusterSetLabel     = "cluster.open-cluster-management.io/clusterset"
+	serviceAccountLabel = "cluster.open-cluster-management.io/submariner-cluster-sa"
+)
+
 const (
 	agentFinalizer            = "cluster.open-cluster-management.io/submariner-agent-cleanup"
-	clusterSetLabel           = "cluster.open-cluster-management.io/clusterset"
-	manifestWorkName          = "submariner-operator"
-	serviceAccountLabel       = "cluster.open-cluster-management.io/submariner-cluster-sa"
+	addOnFinalizer            = "submarineraddon.open-cluster-management.io/submariner-addon-cleanup"
 	submarinerConfigFinalizer = "submarineraddon.open-cluster-management.io/config-cleanup"
-	submarinerLabel           = "cluster.open-cluster-management.io/submariner-agent"
 )
 
 var clusterRBACFiles = []string{
@@ -64,7 +73,6 @@ var sccFiles = []string{
 }
 
 var operatorFiles = []string{
-	"manifests/agent/operator/submariner-operator-namespace.yaml",
 	"manifests/agent/operator/submariner-operator-group.yaml",
 	"manifests/agent/operator/submariner-operator-subscription.yaml",
 	"manifests/agent/operator/submariner.io-submariners-cr.yaml",
@@ -83,9 +91,12 @@ type submarinerAgentController struct {
 	clusterClient      clusterclient.Interface
 	manifestWorkClient workclient.Interface
 	configClient       configclient.Interface
+	addOnClient        addonclient.Interface
 	clusterLister      clusterlisterv1.ManagedClusterLister
 	clusterSetLister   clusterlisterv1alpha1.ManagedClusterSetLister
 	manifestWorkLister worklister.ManifestWorkLister
+	configLister       configlister.SubmarinerConfigLister
+	addOnLister        addonlisterv1alpha1.ManagedClusterAddOnLister
 	eventRecorder      events.Recorder
 }
 
@@ -96,10 +107,12 @@ func NewSubmarinerAgentController(
 	clusterClient clusterclient.Interface,
 	manifestWorkClient workclient.Interface,
 	configClient configclient.Interface,
+	addOnClient addonclient.Interface,
 	clusterInformer clusterinformerv1.ManagedClusterInformer,
 	clusterSetInformer clusterinformerv1alpha1.ManagedClusterSetInformer,
 	manifestWorkInformer workinformer.ManifestWorkInformer,
 	configInformer configinformer.SubmarinerConfigInformer,
+	addOnInformer addoninformerv1alpha1.ManagedClusterAddOnInformer,
 	recorder events.Recorder) factory.Controller {
 	c := &submarinerAgentController{
 		kubeClient:         kubeClient,
@@ -107,9 +120,12 @@ func NewSubmarinerAgentController(
 		clusterClient:      clusterClient,
 		manifestWorkClient: manifestWorkClient,
 		configClient:       configClient,
+		addOnClient:        addOnClient,
 		clusterLister:      clusterInformer.Lister(),
 		clusterSetLister:   clusterSetInformer.Lister(),
 		manifestWorkLister: manifestWorkInformer.Lister(),
+		configLister:       configInformer.Lister(),
+		addOnLister:        addOnInformer.Lister(),
 		eventRecorder:      recorder.WithComponentSuffix("submariner-agent-controller"),
 	}
 	return factory.New().
@@ -118,13 +134,30 @@ func NewSubmarinerAgentController(
 			return accessor.GetName()
 		}, clusterInformer.Informer()).
 		WithInformersQueueKeyFunc(func(obj runtime.Object) string {
+			// TODO: we may consider to use addon to deploy the submariner on the managed cluster instead of
+			// using manifestwork, one problem should be considered - how to get the IPSECPSK
 			accessor, _ := meta.Accessor(obj)
+			if accessor.GetName() != manifestWorkName {
+				return ""
+			}
 			return accessor.GetNamespace()
 		}, manifestWorkInformer.Informer()).
 		WithInformersQueueKeyFunc(func(obj runtime.Object) string {
-			key, _ := cache.MetaNamespaceKeyFunc(obj)
-			return key
+			// TODO: we may consider to use addon to set up the submariner env on the managed cluster instead of
+			// using manifestwork, one problem should be considered - how to get the cloud credentials
+			accessor, _ := meta.Accessor(obj)
+			if accessor.GetName() != helpers.SubmarinerConfigName {
+				return ""
+			}
+			return accessor.GetNamespace() + "/" + accessor.GetName()
 		}, configInformer.Informer()).
+		WithInformersQueueKeyFunc(func(obj runtime.Object) string {
+			accessor, _ := meta.Accessor(obj)
+			if accessor.GetName() != helpers.SubmarinerAddOnName {
+				return ""
+			}
+			return accessor.GetNamespace()
+		}, addOnInformer.Informer()).
 		WithInformers(clusterSetInformer.Informer()).
 		WithSync(c.sync).
 		ToController("SubmarinerAgentController", recorder)
@@ -133,23 +166,28 @@ func NewSubmarinerAgentController(
 func (c *submarinerAgentController) sync(ctx context.Context, syncCtx factory.SyncContext) error {
 	key := syncCtx.QueueKey()
 
-	klog.V(4).Infof("Submariner agent controller is reconciling, queue key: %s", key)
-
 	// if the sync is triggered by change of ManagedClusterSet, reconcile all managed clusters
-	if key == "key" {
-		if err := c.syncAllManagedClusters(ctx); err != nil {
+	if key == factory.DefaultQueueKey {
+		managedClusters, err := c.clusterLister.List(labels.Everything())
+		if err != nil {
 			return err
 		}
+		for _, managedCluster := range managedClusters {
+			// enqueue the managed cluster to reconcile
+			syncCtx.Queue().Add(managedCluster.Name)
+		}
 		return nil
 	}
 
+	klog.V(4).Infof("Submariner agent controller is reconciling, queue key: %s", key)
+
 	namespace, name, err := cache.SplitMetaNamespaceKey(key)
 	if err != nil {
 		// ignore bad format key
 		return nil
 	}
 
-	// if the sync is triggered by change of ManagedCluster or ManifestWork, reconcile the managed cluster
+	// if the sync is triggered by change of ManagedCluster, ManifestWork or ManagedClusterAddOn, reconcile the managed cluster
 	if namespace == "" {
 		managedCluster, err := c.clusterLister.Get(name)
 		if errors.IsNotFound(err) {
@@ -160,26 +198,25 @@ func (c *submarinerAgentController) sync(ctx context.Context, syncCtx factory.Sy
 			return err
 		}
 
-		config, err := c.getSubmarinerConfig(ctx, name)
+		config, err := c.configLister.SubmarinerConfigs(name).Get(helpers.SubmarinerConfigName)
+		if errors.IsNotFound(err) {
+			// only sync the managed cluster
+			return c.syncManagedCluster(ctx, managedCluster.DeepCopy(), nil)
+		}
 		if err != nil {
 			return err
 		}
 
-		if err := c.syncManagedCluster(ctx, managedCluster, config); err != nil {
+		if err := c.syncManagedCluster(ctx, managedCluster.DeepCopy(), config.DeepCopy()); err != nil {
 			return err
 		}
 
-		if config == nil {
-			// there is no submariner configuration, do nothing.
-			return nil
-		}
-
 		// handle creating submariner config before creating managed cluster
-		return c.syncSubmarinerConfig(ctx, managedCluster, config)
+		return c.syncSubmarinerConfig(ctx, managedCluster.DeepCopy(), config.DeepCopy())
 	}
 
 	// if the sync is triggered by change of SubmarinerConfig, reconcile the submariner config
-	config, err := c.configClient.SubmarineraddonV1alpha1().SubmarinerConfigs(namespace).Get(ctx, name, metav1.GetOptions{})
+	config, err := c.configLister.SubmarinerConfigs(namespace).Get(name)
 	if errors.IsNotFound(err) {
 		// config is not found, could have been deleted, do nothing.
 		return nil
@@ -191,39 +228,18 @@ func (c *submarinerAgentController) sync(ctx context.Context, syncCtx factory.Sy
 	managedCluster, err := c.clusterLister.Get(namespace)
 	if errors.IsNotFound(err) {
 		// handle deleting submariner config after managed cluster was deleted.
-		return c.syncSubmarinerConfig(ctx, nil, config)
+		return c.syncSubmarinerConfig(ctx, nil, config.DeepCopy())
 	}
 	if err != nil {
 		return err
 	}
 
 	// the submariner agent config maybe need to update.
-	if err := c.syncManagedCluster(ctx, managedCluster, config); err != nil {
+	if err := c.syncManagedCluster(ctx, managedCluster.DeepCopy(), config.DeepCopy()); err != nil {
 		return err
 	}
 
-	return c.syncSubmarinerConfig(ctx, managedCluster, config)
-}
-
-// syncAllManagedClusters syncs all managed clusters
-func (c *submarinerAgentController) syncAllManagedClusters(ctx context.Context) error {
-	managedClusters, err := c.clusterLister.List(labels.Everything())
-	if err != nil {
-		return err
-	}
-
-	errs := []error{}
-	for _, managedCluster := range managedClusters {
-		config, err := c.getSubmarinerConfig(ctx, managedCluster.ClusterName)
-		if err != nil {
-			errs = append(errs, err)
-		}
-		if err = c.syncManagedCluster(ctx, managedCluster, config); err != nil {
-			errs = append(errs, err)
-		}
-	}
-
-	return operatorhelpers.NewMultiLineAggregate(errs)
+	return c.syncSubmarinerConfig(ctx, managedCluster.DeepCopy(), config.DeepCopy())
 }
 
 // syncManagedCluster syncs one managed cluster
@@ -231,23 +247,12 @@ func (c *submarinerAgentController) syncManagedCluster(
 	ctx context.Context,
 	managedCluster *clusterv1.ManagedCluster,
 	config *configv1alpha1.SubmarinerConfig) error {
-	// the cluster does not have the submariner label, try to clean up the submariner agent
-	if _, existed := managedCluster.Labels[submarinerLabel]; !existed {
-		return c.cleanUpSubmarinerAgent(ctx, managedCluster)
-	}
-
-	// the cluster does not have the clusterset label, try to clean up the submariner agent
-	clusterSetName, existed := managedCluster.Labels[clusterSetLabel]
-	if !existed {
-		return c.cleanUpSubmarinerAgent(ctx, managedCluster)
-	}
-
-	// find the clustersets that contains this managed cluster
-	// if the clusterset is not found, try to clean up the submariner agent
-	_, err := c.clusterClient.ClusterV1alpha1().ManagedClusterSets().Get(ctx, clusterSetName, metav1.GetOptions{})
+	// find the submariner-addon on the managed cluster namespace
+	addOn, err := c.addOnLister.ManagedClusterAddOns(managedCluster.Name).Get(helpers.SubmarinerAddOnName)
 	switch {
 	case errors.IsNotFound(err):
-		return c.cleanUpSubmarinerAgent(ctx, managedCluster)
+		// submariner-addon is not found, could have been deleted, do nothing.
+		return nil
 	case err != nil:
 		return err
 	}
@@ -270,16 +275,55 @@ func (c *submarinerAgentController) syncManagedCluster(
 
 	// managed cluster is deleting, we remove its related resources
 	if !managedCluster.DeletionTimestamp.IsZero() {
-		return c.cleanUpSubmarinerAgent(ctx, managedCluster)
+		return c.cleanUpSubmarinerAgent(ctx, managedCluster, addOn)
 	}
 
-	return c.deploySubmarinerAgent(ctx, clusterSetName, managedCluster, config)
+	clusterSetName, existed := managedCluster.Labels[clusterSetLabel]
+	if !existed {
+		// the cluster does not have the clusterset label, try to clean up the submariner agent
+		return c.cleanUpSubmarinerAgent(ctx, managedCluster, addOn)
+	}
+
+	// find the clustersets that contains this managed cluster
+	_, err = c.clusterSetLister.Get(clusterSetName)
+	switch {
+	case errors.IsNotFound(err):
+		// if one cluster has clusterset label, but the clusterset is not found, it could have been deleted
+		// try to clean up the submariner agent
+		return c.cleanUpSubmarinerAgent(ctx, managedCluster, addOn)
+	case err != nil:
+		return err
+	}
+
+	// add a finalizer to the submariner-addon
+	if addOn.DeletionTimestamp.IsZero() {
+		hasFinalizer := false
+		for i := range addOn.Finalizers {
+			if addOn.Finalizers[i] == addOnFinalizer {
+				hasFinalizer = true
+				break
+			}
+		}
+		if !hasFinalizer {
+			addOn.Finalizers = append(addOn.Finalizers, addOnFinalizer)
+			_, err := c.addOnClient.AddonV1alpha1().ManagedClusterAddOns(managedCluster.Name).Update(ctx, addOn, metav1.UpdateOptions{})
+			return err
+		}
+	}
+
+	// submariner-addon is deleting, we remove its related resources
+	if !addOn.DeletionTimestamp.IsZero() {
+		return c.cleanUpSubmarinerAgent(ctx, managedCluster, addOn)
+	}
+
+	return c.deploySubmarinerAgent(ctx, clusterSetName, managedCluster, addOn, config)
 }
 
 // syncSubmarinerConfig syncs submariner configuration
 func (c *submarinerAgentController) syncSubmarinerConfig(ctx context.Context,
 	managedCluster *clusterv1.ManagedCluster,
 	config *configv1alpha1.SubmarinerConfig) error {
+	// add a finalizer to the submarinerconfig
 	if config.DeletionTimestamp.IsZero() {
 		hasFinalizer := false
 		for i := range config.Finalizers {
@@ -300,20 +344,29 @@ func (c *submarinerAgentController) syncSubmarinerConfig(ctx context.Context,
 		if err := c.cleanUpSubmarinerClusterEnv(ctx, config); err != nil {
 			return err
 		}
-		return c.removeConfigFinalizer(ctx, config)
-	}
-
-	if config.Spec.CredentialsSecret == nil {
-		// no platform credentials, the submariner cluster environment neet not to be prepared
-		return nil
+		return helpers.RemoveConfigFinalizer(ctx, c.configClient, config, submarinerConfigFinalizer)
 	}
 
 	if managedCluster == nil {
 		return nil
 	}
-
 	managedClusterInfo := helpers.GetManagedClusterInfo(managedCluster)
 
+	if config.Spec.CredentialsSecret == nil {
+		// no platform credentials, the cluster env is not requred to prepare, only update the manged cluster info
+		_, _, err := helpers.UpdateSubmarinerConfigStatus(
+			c.configClient,
+			config.Namespace, config.Name,
+			helpers.UpdateSubmarinerConfigStatusFn(metav1.Condition{
+				Type:    configv1alpha1.SubmarinerConfigConditionApplied,
+				Status:  metav1.ConditionTrue,
+				Reason:  "SubmarinerConfigApplied",
+				Message: "SubmarinerConfig was applied",
+			}, managedClusterInfo),
+		)
+		return err
+	}
+
 	// prepare submariner cluster environment
 	errs := []error{}
 	cloudProvider, preparedErr := cloud.GetCloudProvider(c.kubeClient, c.manifestWorkClient, c.eventRecorder, managedClusterInfo, config)
@@ -351,11 +404,16 @@ func (c *submarinerAgentController) syncSubmarinerConfig(ctx context.Context,
 }
 
 // clean up the submariner agent from this managedCluster
-func (c *submarinerAgentController) cleanUpSubmarinerAgent(ctx context.Context, managedCluster *clusterv1.ManagedCluster) error {
+func (c *submarinerAgentController) cleanUpSubmarinerAgent(ctx context.Context, managedCluster *clusterv1.ManagedCluster, addOn *addonv1alpha1.ManagedClusterAddOn) error {
 	if err := c.removeSubmarinerAgent(ctx, managedCluster.Name); err != nil {
 		return err
 	}
-	return c.removeAgentFinalizer(ctx, managedCluster)
+
+	if err := c.removeAgentFinalizer(ctx, managedCluster); err != nil {
+		return err
+	}
+
+	return helpers.RemoveAddOnFinalizer(ctx, c.addOnClient, addOn, addOnFinalizer)
 }
 
 // removeAgentFinalizer removes the agent finalizer from a clusterset
@@ -381,9 +439,10 @@ func (c *submarinerAgentController) deploySubmarinerAgent(
 	ctx context.Context,
 	clusterSetName string,
 	managedCluster *clusterv1.ManagedCluster,
+	managedClusterAddOn *addonv1alpha1.ManagedClusterAddOn,
 	submarinerConfig *configv1alpha1.SubmarinerConfig) error {
 	// generate service account and bind it to `submariner-k8s-broker-cluster` role
-	brokerNamespace := fmt.Sprintf("submariner-clusterset-%s-broker", clusterSetName)
+	brokerNamespace := helpers.GernerateBrokerName(clusterSetName)
 	if err := c.applyClusterRBACFiles(brokerNamespace, managedCluster.Name); err != nil {
 		return err
 	}
@@ -397,6 +456,7 @@ func (c *submarinerAgentController) deploySubmarinerAgent(
 		managedCluster,
 		brokerNamespace,
 		submarinerConfig,
+		managedClusterAddOn,
 	)
 	if err != nil {
 		return fmt.Errorf("failed to create submariner brokerInfo of cluster %v : %v", managedCluster.Name, err)
@@ -407,11 +467,10 @@ func (c *submarinerAgentController) deploySubmarinerAgent(
 	if err != nil {
 		return err
 	}
-	if err := helpers.ApplyManifestWork(ctx, c.manifestWorkClient, operatorManifestWork); err != nil {
+	if err := helpers.ApplyManifestWork(ctx, c.manifestWorkClient, operatorManifestWork, c.eventRecorder); err != nil {
 		return err
 	}
 
-	c.eventRecorder.Event("SubmarinerAgentDeployed", fmt.Sprintf("submariner agent was deployed on managed cluster %q", managedCluster.Name))
 	return nil
 }
 
@@ -419,10 +478,14 @@ func (c *submarinerAgentController) removeSubmarinerAgent(ctx context.Context, c
 	errs := []error{}
 	// remove submariner manifestworks
 	err := c.manifestWorkClient.WorkV1().ManifestWorks(clusterName).Delete(ctx, manifestWorkName, metav1.DeleteOptions{})
-	if err != nil && !errors.IsNotFound(err) {
+	switch {
+	case errors.IsNotFound(err):
+		//there is no submariner manifestworks, do noting
+	case err == nil:
+		c.eventRecorder.Eventf("SubmarinerManifestWorksDeleted", "Deleted manifestwork %q", fmt.Sprintf("%s/%s", clusterName, manifestWorkName))
+	case err != nil:
 		errs = append(errs, fmt.Errorf("failed to remove submariner agent from managed cluster %v: %v", clusterName, err))
 	}
-	c.eventRecorder.Eventf("SubmarinerManifestWorksDeleted", "Deleted manifestwork %q", fmt.Sprintf("%s/%s", clusterName, manifestWorkName))
 
 	// remove service account and its rolebinding from broker namespace
 	if err := c.removeClusterRBACFiles(ctx, clusterName); err != nil {
@@ -488,23 +551,6 @@ func (c *submarinerAgentController) removeClusterRBACFiles(ctx context.Context,
 	)
 }
 
-func (c *submarinerAgentController) getSubmarinerConfig(ctx context.Context, namespace string) (*configv1alpha1.SubmarinerConfig, error) {
-	configs, err := c.configClient.SubmarineraddonV1alpha1().SubmarinerConfigs(namespace).List(ctx, metav1.ListOptions{})
-	if err != nil {
-		return nil, err
-	}
-	switch len(configs.Items) {
-	case 0:
-		return nil, nil
-	case 1:
-		return &configs.Items[0], nil
-	default:
-		//TODO we need ensure only one config for one managed cluster in the futrue
-		c.eventRecorder.Warningf("one more than submariner configs are found from %q", namespace)
-		return nil, nil
-	}
-}
-
 func (c *submarinerAgentController) cleanUpSubmarinerClusterEnv(ctx context.Context, config *configv1alpha1.SubmarinerConfig) error {
 	// no platform credentials, the submariner cluster environment is not prepared
 	if config.Spec.CredentialsSecret == nil {
@@ -527,27 +573,9 @@ func (c *submarinerAgentController) cleanUpSubmarinerClusterEnv(ctx context.Cont
 	return nil
 }
 
-func (c *submarinerAgentController) removeConfigFinalizer(ctx context.Context, config *configv1alpha1.SubmarinerConfig) error {
-	copiedFinalizers := []string{}
-	for i := range config.Finalizers {
-		if config.Finalizers[i] == submarinerConfigFinalizer {
-			continue
-		}
-		copiedFinalizers = append(copiedFinalizers, config.Finalizers[i])
-	}
-
-	if len(config.Finalizers) != len(copiedFinalizers) {
-		config.Finalizers = copiedFinalizers
-		_, err := c.configClient.SubmarineraddonV1alpha1().SubmarinerConfigs(config.Namespace).Update(ctx, config, metav1.UpdateOptions{})
-		return err
-	}
-
-	return nil
-}
-
 func getManifestWork(managedCluster *clusterv1.ManagedCluster, config interface{}) (*workv1.ManifestWork, error) {
 	files := []string{agentRBACFile}
-	if helpers.GetClusterType(managedCluster) == helpers.ClusterTypeOCP {
+	if helpers.GetClusterProduct(managedCluster) == helpers.ProductOCP {
 		files = append(files, sccFiles...)
 	}
 	files = append(files, operatorFiles...)
